import pandas as pd 
import numpy as np
import json
from .utils import _extract_attributes, convert_df_to_trajectory_list


if __name__ == '__main__':
    NUM_TRAJ = 20000
    event_times = np.random.gamma(shape=200, scale=1, size=NUM_TRAJ)

    # Generate trajectories

    trajectory_dfs = []
    for i in range(NUM_TRAJ):
        # Generate the trajectory length
        traj_len = np.random.randint(low=1, high=3000)
        # Generate the times
        pred_times = np.sort(np.random.uniform(low=0, high=event_times[i], size=traj_len))
        # Generate the risk scores
        pred_risks = np.random.uniform(low=0, high=1, size=traj_len)
        # Generate true or false for whether the event occurred or not
        event_occurred = np.random.rand() > 0.5
        
        traj_df = pd.DataFrame({
            "id": i,
            "event_occurred": event_occurred,
            "event_time": event_times[i],
            "predicted_times": pred_times,
            "predicted_risks": pred_risks
        })
        
        trajectory_dfs.append(traj_df)

    trajectory_df = pd.concat(trajectory_dfs)
    traj_list = convert_df_to_trajectory_list(trajectory_df, 'id', 'event_occurred', 'event_time', 'predicted_times', 'predicted_risks')

    with open("../benchmark_data/benchmark_trajectories_200.json", "w") as f:
        json.dump(traj_list[0:200], f, indent=2)

    with open("../benchmark_data/benchmark_trajectories_2000.json", "w") as f:
        json.dump(traj_list[0:2000], f, indent=2)
        
    with open("../benchmark_data/benchmark_trajectories_20000.json", "w") as f:
        json.dump(traj_list, f, indent=2)


from .utils import validate_trajectories_schema
def get_valid_times(predicted_times, predicted_risks, threshold, snooze_window):

    result = []
    snooze_boundary = -1.0
    for time, risk in zip(predicted_times, predicted_risks):
        if time > snooze_boundary and risk > threshold:
            result.append(time)
            snooze_boundary = time + snooze_window
    return result

def get_prediction_level_metrics(positive_prediction_times, detection_window, event_time):
    num_tp = sum(1 for p in positive_prediction_times if event_time - detection_window <= p < event_time)
    num_fp = len(positive_prediction_times) - num_tp
    return num_tp, num_fp

def calculate_metrics(trajectory_list, threshold, snooze_window, detection_window):
    episode_tp = episode_fp = episode_fn = episode_tn = 0
    prediction_tp = prediction_fp = 0

    for traj in trajectory_list:
        positive_predictions = get_valid_times(traj['predicted_times'], traj['predicted_risks'], threshold, snooze_window)
        if traj['event_occurred']:
            num_tp, num_fp = get_prediction_level_metrics(positive_predictions, detection_window, traj['event_time'])
            episode_tp += (num_tp > 0)
            episode_fn += (num_tp == 0)
            prediction_tp += num_tp
            prediction_fp += num_fp
        else:
            episode_fp += (len(positive_predictions) > 0)
            episode_tn += (len(positive_predictions) == 0)
            prediction_fp += len(positive_predictions)

    return {
        'threshold': threshold,
        'episode_tp': episode_tp,
        'episode_fp': episode_fp,
        'episode_tn': episode_tn,
        'episode_fn': episode_fn,
        'prediction_tp': prediction_tp,
        'prediction_fp': prediction_fp
    }

def get_metrics(trajectory_list, snooze_window, detection_window):
    validate_trajectories_schema(trajectory_list)  # Validate trajectories
    if snooze_window == 0:
        return get_metrics_no_snooze(trajectory_list, detection_window)
    
    risk_scores = sorted({risk for traj in trajectory_list for risk in traj['predicted_risks']}, reverse=True)
    return [calculate_metrics(trajectory_list, th, snooze_window, detection_window) for th in risk_scores]

def get_approximate_metrics(trajectory_list, thresholds, snooze_window, detection_window):
    validate_trajectories_schema(trajectory_list)  # Validate trajectories
    if snooze_window == 0:
        print("Warning: Snooze window is set to 0. Using get_metrics instead for better performance and accuracy.")
        return get_metrics(trajectory_list, snooze_window, detection_window)
    return [calculate_metrics(trajectory_list, th, snooze_window, detection_window) for th in thresholds]

def get_metrics_no_snooze(trajectories, detection_window=24):
    validate_trajectories_schema(trajectories)  # Validate trajectories
    event_occurs_episodes = set()
    all_episodes = set()
    risk_scores = {}

    for i, traj in enumerate(trajectories):
        traj['id'] = i
        all_episodes.add(i)
        if traj['event_occurred']:
            event_occurs_episodes.add(i)
            within_detection_window = [(traj['event_time'] - t) <= detection_window for t in traj['predicted_times']]
        else:
            within_detection_window = [False] * len(traj['predicted_times'])
        
        for risk, within_window in zip(traj['predicted_risks'], within_detection_window):
            risk_scores.setdefault(risk, []).append((within_window, i))

    sorted_risk_scores = sorted(risk_scores.keys(), reverse=True)
    positive_prediction_episodes = set()
    negative_prediction_episodes = set()
    episode_tp = episode_fp = episode_fn = episode_tn = 0
    prediction_tp = prediction_fp = 0
    
    output_metrics = []
    for score in sorted_risk_scores:
        for det_window, ep_id in risk_scores[score]:
            if det_window:
                prediction_tp += 1
                if ep_id not in positive_prediction_episodes:
                    positive_prediction_episodes.add(ep_id)
                    episode_tp += (ep_id in event_occurs_episodes)
            else:
                prediction_fp += 1
                if ep_id not in negative_prediction_episodes:
                    negative_prediction_episodes.add(ep_id)
                    episode_fp += (ep_id not in event_occurs_episodes)

        episode_fn = len(event_occurs_episodes) - episode_tp
        episode_tn = len(all_episodes) - len(event_occurs_episodes) - episode_fp

        output_metrics.append({
            "threshold": score,
            "episode_tp": episode_tp,
            "episode_fp": episode_fp,
            "episode_tn": episode_tn,
            "episode_fn": episode_fn,
            "prediction_tp": prediction_tp,
            "prediction_fp": prediction_fp
        })

    return output_metrics



from typing import List, Dict, Tuple
import pandas as pd

def _recall(metric: Dict, total_positive_episodes: float) -> float:
    """
    Calculate recall for a single metric.
    
    Args:
        metric: A dictionary (obtained from compute_metrics) containing 'episode_tp'.
        total_positive_episodes: The total number of positive episodes.
    Returns:
        The recall value.
    """
    return metric['episode_tp'] / total_positive_episodes

def _precision(metric: Dict) -> float:
    """
    Calculate precision for a single metric.
    
    Args:
        metric: A dictionary (obtained from compute_metrics) containing 'prediction_tp' and 'prediction_fp'.
    Returns:
        The precision value or None if division by zero occurs.
    """
    try:
        return metric['prediction_tp'] / (metric['prediction_tp'] + metric['prediction_fp'])
    except ZeroDivisionError:
        return None

def _fpr(metric: Dict, total_negative_episodes: float) -> float:
    """
    Calculate false positive rate for a single metric.

    Args:
        metric: A dictionary (obtained from compute_metrics) containing 'episode_fp' and 'episode_tn'.
        total_negative_episodes: The total number of negative episodes.
    Returns:
        The false positive rate.
    """
    return metric['episode_fp'] / total_negative_episodes

def validate_trajectories_schema(trajectories: List[Dict]):
    """
    Validate the schema of a list of trajectory dictionaries.

    Args:
        trajectories: A list of trajectory dictionaries.
    Raises:
        ValueError: If the schema is invalid.
    """
    required_keys = {'predicted_times', 'predicted_risks', 'event_occurred', 'event_time'}
    
    for traj in trajectories:
        if not isinstance(traj, Dict):
            raise ValueError("Each trajectory must be a dictionary.")
        
        if not required_keys.issubset(traj.keys()):
            raise ValueError(f"Each trajectory must contain the keys: {required_keys}")
        
        if not isinstance(traj['predicted_times'], List) or not all(isinstance(t, (int, float)) for t in traj['predicted_times']):
            raise ValueError("The 'predicted_times' key must be a list of numbers.")
        
        if len(traj['predicted_times']) < 1:
            raise ValueError("The 'predicted_times' list must contain at least one element.")
        
        if not isinstance(traj['predicted_risks'], List) or not all(isinstance(r, (int, float)) for r in traj['predicted_risks']):
            raise ValueError("The 'predicted_risks' key must be a list of numbers.")
        
        if len(traj['predicted_risks']) < 1:
            raise ValueError("The 'predicted_risks' list must contain at least one element.")
        
        if not isinstance(traj['event_occurred'], bool):
            raise ValueError("The 'event_occurred' key must be a boolean.")
        
        if not isinstance(traj['event_time'], (int, float)):
            raise ValueError("The 'event_time' key must be a number.")
        
        # Coerce predicted_times, predicted_risks, and event_time to floats
        traj['predicted_times'] = [float(t) for t in traj['predicted_times']]
        traj['predicted_risks'] = [float(r) for r in traj['predicted_risks']]
        traj['event_time'] = float(traj['event_time'])
        
    return True

def auprc_score(metrics: List[Dict]) -> float:
    """
    Compute the area under the precision-recall curve, using the lower trapezoidal estimator, which is a point-estimator for the area under the precision-recall curve.
    This is chosen due to its performance as an estimator as described in Boyd(2013) and its ability to be computed even when the precision-recall is not necessarily monotonic.

    Args:
        metrics: A list of metric dictionaries.
    Returns:
        The average precision score.
    """
    recall_precision_dict = {}
    total_positive_episodes = metrics[-1]['episode_tp'] + metrics[-1]['episode_fn']
    for m in metrics:
        # Compute recall value
        r = _recall(m, total_positive_episodes)
        
        # Compute precision value
        p = _precision(m)

        if p is not None:
            if r not in recall_precision_dict:
                recall_precision_dict[r] = {'pmin': p, 'pmax': p}
            else:
                recall_precision_dict[r]['pmin'] = min(recall_precision_dict[r]['pmin'], p)
                recall_precision_dict[r]['pmax'] = max(recall_precision_dict[r]['pmax'], p)
    # Sort by recall
    sorted_recalls = sorted(recall_precision_dict.keys())
    precisions_min = [recall_precision_dict[r]['pmin'] for r in sorted_recalls]
    precisions_max = [recall_precision_dict[r]['pmax'] for r in sorted_recalls]

    ap_sum = 0
    for i in range(0, len(sorted_recalls)-1):
        ap_sum += (precisions_min[i] + precisions_max[i+1])/ 2 * (sorted_recalls[i+1] - sorted_recalls[i])

    return ap_sum

def precision_recall_curve(metrics: List[Dict]) -> Tuple[List[float], List[float]]:
    """
    Generate the precision-recall curve data. Note that this may alias sklearn.metrics.precision_recall_curve.

    Args:
        metrics: A list of metric dictionaries.
    Returns:
        A tuple containing two lists: recalls and precisions.
    """
    xs = []
    ys = []
    total_positive_episodes = metrics[-1]['episode_tp'] + metrics[-1]['episode_fn']
    
    for m in metrics:
        curr_prec = _precision(m)
        if curr_prec is not None:
            curr_recall = _recall(m, total_positive_episodes)
            xs.append(curr_recall)
            ys.append(curr_prec)
    return xs, ys

def recall_at_fixed_precision(metrics: List[Dict], target_precision: float) -> float:
    """
    Find the closest recall value at the specified precision.

    Args:
        metrics: List of dictionaries containing 'episode_tp', 'episode_fp', 'episode_fn', and 'prediction_tp'.
        target_precision: The precision value to find the closest recall for.
    Returns:
        The recall value closest to the specified precision, or None if not found.
    """
    total_positive_episodes = metrics[-1]['episode_tp'] + metrics[-1]['episode_fn']

    closest_recall = None
    smallest_diff = float('inf')

    for m in metrics:
        curr_prec = _precision(m)
        if curr_prec is not None:
            diff = abs(curr_prec - target_precision)
            if diff < smallest_diff:
                smallest_diff = diff
                closest_recall = _recall(m, total_positive_episodes)

    return closest_recall

def precision_at_fixed_recall(metrics: List[Dict], target_recall: float) -> float:
    """
    Find the closest precision value at the specified recall.

    Args:
        metrics: List of dictionaries containing 'episode_tp', 'episode_fp', 'episode_fn', and 'prediction_tp'.
        target_recall: The recall value to find the closest precision for.
    Returns:
        The precision value closest to the specified recall, or None if not found.
    """
    total_positive_episodes = metrics[-1]['episode_tp'] + metrics[-1]['episode_fn']

    closest_precision = None
    smallest_diff = float('inf')

    for m in metrics:
        curr_prec = _precision(m)
        if curr_prec is not None:
            curr_recall = _recall(m, total_positive_episodes)
            diff = abs(curr_recall - target_recall)
            if diff < smallest_diff:
                smallest_diff = diff
                closest_precision = curr_prec

    return closest_precision

def roc_curve(metrics: List[Dict]) -> Tuple[List[float], List[float]]:
    """
    Generate the ROC curve data.

    Args:
        metrics: A list of metric dictionaries. Note that this may alias sklearn.metrics.roc_curve.
    Returns:
        A tuple containing two lists: false positive rates and true positive rates.
    """
    fprs = []
    tprs = []
    total_positive_episodes = metrics[-1]['episode_tp'] + metrics[-1]['episode_fn']
    total_negative_episodes = metrics[-1]['episode_fp'] + metrics[-1].get('episode_tn', 0)

    for m in metrics:
        curr_tpr = _recall(m, total_positive_episodes)
        curr_fpr = _fpr(m, total_negative_episodes)
        fprs.append(curr_fpr)
        tprs.append(curr_tpr)
    return fprs, tprs

def _calculate_trapezoid_area(x1: float, x2: float, y1: float, y2: float) -> float:
    """
    Calculate the area of a trapezoid.

    Args:
        x1: x-coordinate of the first point
        x2: x-coordinate of the second point
        y1: y-coordinate of the first point
        y2: y-coordinate of the second point
    Returns:
        Area of the trapezoid
    """
    return 0.5 * (x2 - x1) * (y1 + y2)

def auroc_score(metrics: List[Dict]) -> float:
    """
    Compute the Area Under the ROC Curve (AUROC).

    Args:
        metrics: A list of metric dictionaries.
    Returns:
        The AUROC value.
    """
    fprs, tprs = roc_curve(metrics)
    # Sort the FPR and TPR for proper integration
    sorted_pairs = sorted(zip(fprs, tprs))
    sorted_fprs, sorted_tprs = zip(*sorted_pairs)

    # Compute the Area Under the Curve using the trapezoidal rule
    auroc = 0.0
    for i in range(1, len(sorted_fprs)):
        auroc += _calculate_trapezoid_area(sorted_fprs[i-1], sorted_fprs[i], sorted_tprs[i-1], sorted_tprs[i])

    return auroc

def _extract_attributes(groupbydf: 'pd.core.groupby.generic.DataFrameGroupBy',
                        event_occurred_col: str, 
                        event_time_col: str, 
                        predicted_times_col: str, 
                        predicted_risks_col: str) -> Dict:
    """
    Extract attributes from a grouped dataframe. A helper function used for converting a dataframe to a list of trajectories.

    Args:
        groupbydf: A dataframe grouped by episode_id.
        event_occurred_col: The column name for event occurrence.
        event_time_col: The column name for event time.
        predicted_times_col: The column name for predicted times.
        predicted_risks_col: The column name for predicted risks.
    Returns:
        A dictionary containing the extracted attributes.
    """
    result_dict = {}
    result_dict['event_occurred'] = bool(groupbydf[event_occurred_col].iloc[0])
    result_dict['event_time'] = groupbydf[event_time_col].iloc[0]
    result_dict['predicted_times'] = groupbydf[predicted_times_col].tolist()
    result_dict['predicted_risks'] = groupbydf[predicted_risks_col].tolist()
    return result_dict
    
def convert_df_to_trajectory_list(df: 'pd.DataFrame', 
                                  episode_id_col: str, 
                                  event_occurred_col: str, 
                                  event_time_col: str, 
                                  predicted_times_col: str, 
                                  predicted_risks_col: str) -> List[Dict]:
    """
    Convert a dataframe to a list of trajectory dictionaries.

    Args:
        df: A dataframe.
        episode_id_col: The column name for episode id.
        event_occurred_col: The column name for event occurrence.
        event_time_col: The column name for event time.
        predicted_times_col: The column name for predicted times.
        predicted_risks_col: The column name for predicted risks.
    Returns:
        A list of trajectory dictionaries.
    """
    traj_list = []
    for i, df in df.groupby(episode_id_col):
        traj_list.append(_extract_attributes(df, event_occurred_col, event_time_col, predicted_times_col, predicted_risks_col))
    return traj_list


from functools import wraps
from typing import List, Dict, Optional
from ctseval.utils import (validate_trajectories_schema as _validate_trajectories_schema,
                    auprc_score as _auprc_score,
                    precision_recall_curve as _precision_recall_curve,
                    recall_at_fixed_precision as _recall_at_fixed_precision,
                    precision_at_fixed_recall as _precision_at_fixed_recall,
                    roc_curve as _roc_curve,
                    auroc_score as _auroc_score,
                    _extract_attributes,
                    convert_df_to_trajectory_list as _convert_df_to_trajectory_list)
from ctseval._ctseval import compute_metrics_c

def compute_metrics(trajectories: List[Dict], snooze_window: float, detection_window: float, verbosity: Optional[int] = 1) -> List[Dict]:
    """
    Compute metrics for clinical time series predictions.
    
    This function calculates various metrics for evaluating the performance of clinical time series predictions,
    including episode-level and prediction-level metrics.

    Args:
        trajectories (List[Dict]): List of trajectory dictionaries. Each dictionary should contain:
            - predicted_times (List[float]): List of prediction times.
            - predicted_risks (List[float]): List of predicted risk scores.
            - event_occurred (bool): Whether an event occurred or not.
            - event_time (float): Time of the event, if it occurred.
        snooze_window (float): The snooze window duration in units of time.
        detection_window (float): The detection window duration in units of time.
        verbosity (Optional[int], optional): Verbosity level. Defaults to 1.
    
    Returns:
        List[Dict]: List of dictionaries with episode-level and prediction-level metrics.
            Each dictionary contains the following keys:
            - threshold (float): The risk threshold used for this set of metrics.
            - episode_tp (int): True positive episodes.
            - episode_fp (int): False positive episodes.
            - episode_tn (int): True negative episodes.
            - episode_fn (int): False negative episodes.
            - prediction_tp (int): True positive predictions.
            - prediction_fp (int): False positive predictions.

    Raises:
        ValueError: If the trajectories schema is invalid.

    Example:
        >>> trajectories = [
        ...     {
        ...         "predicted_times": [1.0, 2.0, 3.0],
        ...         "predicted_risks": [0.1, 0.2, 0.3],
        ...         "event_occurred": True,
        ...         "event_time": 2.5
        ...     },
        ...     {
        ...         "predicted_times": [1.0, 2.0, 3.0],
        ...         "predicted_risks": [0.05, 0.15, 0.25],
        ...         "event_occurred": False,
        ...         "event_time": 0
        ...     }
        ... ]
        >>> metrics = compute_metrics(trajectories, snooze_window=1.0, detection_window=1.0)
        >>> print(metrics[0])
        {'threshold': 0.3, 'episode_tp': 1, 'episode_fp': 0, 'episode_tn': 1, 'episode_fn': 0, 'prediction_tp': 1, 'prediction_fp': 0}
    """
    validate_trajectories_schema(trajectories)
    results = compute_metrics_c(trajectories, snooze_window, detection_window, verbosity)
    return results

@wraps(_validate_trajectories_schema)
def validate_trajectories_schema(*args, **kwargs):
    return _validate_trajectories_schema(*args, **kwargs)

@wraps(_auprc_score)
def auprc_score(*args, **kwargs):
    return _auprc_score(*args, **kwargs)

@wraps(_precision_recall_curve)
def precision_recall_curve(*args, **kwargs):
    return _precision_recall_curve(*args, **kwargs)

@wraps(_recall_at_fixed_precision)
def recall_at_fixed_precision(*args, **kwargs):
    return _recall_at_fixed_precision(*args, **kwargs)

@wraps(_precision_at_fixed_recall)
def precision_at_fixed_recall(*args, **kwargs):
    return _precision_at_fixed_recall(*args, **kwargs)

@wraps(_roc_curve)
def roc_curve(*args, **kwargs):
    return _roc_curve(*args, **kwargs)

@wraps(_auroc_score)
def auroc_score(*args, **kwargs):
    return _auroc_score(*args, **kwargs)

@wraps(_extract_attributes)
def extract_attributes(*args, **kwargs):
    return _extract_attributes(*args, **kwargs)

@wraps(_convert_df_to_trajectory_list)
def convert_df_to_trajectory_list(*args, **kwargs):
    return _convert_df_to_trajectory_list(*args, **kwargs)
    

__all__ = ['compute_metrics', 'validate_trajectories_schema', 'auprc_score', 'precision_recall_curve',
           'recall_at_fixed_precision', 'precision_at_fixed_recall', 'roc_curve', 'auroc_score',
           'extract_attributes', 'convert_df_to_trajectory_list']


from setuptools import setup, Extension, find_packages
import numpy as np

# Define the extension module
ctseval_module = Extension(
    'ctseval._ctseval',
    sources=[
        'ctseval/ctseval.c',
        'ctseval/_ctseval.c',
    ],
    include_dirs=[np.get_include(), 'ctseval/include'],
)

# Setup the package
setup(
    name='ctseval',
    version='0.1.0',
    packages=find_packages(),
    ext_modules=[ctseval_module],
    package_data={
        'ctseval': ['include/*.h'],
    },
    include_package_data=True,
    author='Michael Gao',
    author_email='michaelgao8@gmail.com',
    description='A package for evaluating clinical time series predictions',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/michaelgao8/ctseval',
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.6',
    install_requires=[
        'numpy',
        'pandas',
    ],
)